defmodule GiveAway.Channel do
  @moduledoc """
  Contains the API for all social media channel modules
  """

  @type url :: String.t

  @type param_key :: String.t

  @type param_value :: String.t

  @type param_tuple :: {param_key, param_value}

  defmacro __using__(_) do
    quote do
      @behaviour GiveAway.Channel

      def url(params) do
        base_url()
        |> URI.parse
        |> put_query(prepare_params(params))
        |> URI.to_string
      end

      defp prepare_params(params) do
        params
        |> Stream.map(&parse_param/1)
        |> Enum.filter(&valid_param?/1)
      end

      defp put_query(uri, params) do
        old_params = URI.decode_query(uri.query || "")
        new_params = Enum.into(params, %{})
        merged_params = Map.merge(old_params, new_params)

        %URI{uri | query: URI.encode_query(merged_params)}
      end

      defp valid_param?({key, _}) do
        key in valid_params()
      end
    end
  end

  @callback valid_params() :: [param_key]
  @callback parse_param(param_tuple) :: param_tuple
  @callback base_url() :: url
end
